.olLayerGoogleCopyright {
    right: 3px;
    bottom: 2px;
    left: auto;  
}
.olLayerGoogleV3.olLayerGoogleCopyright {
    bottom: 0px;
    right: 0px !important;
}
.olLayerGooglePoweredBy {
    left: 2px;
    bottom: 2px;   
}
.olLayerGoogleV3.olLayerGooglePoweredBy {
    bottom: 0px !important;
}

