{{py:
from html import escape
import textwrap

wrapper = textwrap.TextWrapper(replace_whitespace=False, width=90,
                               break_long_words=False)
menu_title = "Capabilities"
jscript_openlayers = None
jscript_functions = None
}}
            <h2>{{service}} GetCapabilities</h2>
            <a href="{{url}}">{{url}}</a>
            <pre>
{{capabilities}}
            </pre>

