{% extends "_base.html" %}
{% block title %}{{ super() }} {% trans %}TileMatrixSets{% endtrans %} {% endblock %}
{% block crumbs %}{{ super() }}
/ <a href="./tileMatrixSets">{% trans %}TileMatrixSets{% endtrans %}</a>
{% endblock %}
{% block body %}
    <section id="tilematrixsets">

      <h2>{% trans %}Tile matrix sets available in this service{% endtrans %}</h2>
      <div class="row">
        <div class="col-sm-12 col-md-6">
          <table class="table table-striped table-bordered">
            <thead>
              <tr>
                <th>{% trans %}Title{% endtrans %}</th>
                <th>{% trans %}Description{% endtrans %}</th>
              </tr>
            </thead>
            <tbody>
              {% for tms in data['tileMatrixSets'] %}
              <tr itemscope itemtype="https://schema.org/WebAPI">
                <td itemprop="name" data-label="title">
                  {% for link in tms['links'] %}
                  {% if (link['rel']=="self" and link['type']=="text/html") %}
                  <a title="{{ tms.id }}" href="{{ link.href }}">{{tms.id}}</a>
                  {% endif %}
                  {% endfor %}
                </td>
                <td itemprop="description" data-label="description">
                  {{ tms.title }}
                </td>
              </tr>
              {% endfor %}
            </tbody>
          </table>
        </div>
      </div>
    </section>
{% endblock %}
