{% extends "_base.html" %}
{% block title %}{{ super() }} {% trans %}Collections{% endtrans %}{% endblock %}
{% block crumbs %}{{ super() }}
/ <a href="{{ data['collections_path'] }}">{% trans %}Collections{% endtrans %}</a>
{% endblock %}
{% block body %}
    <section id="data-collections">
      <h1>{% trans %}Data collections in this service{% endtrans %}</h1>
            <table class="table table-striped table-bordered">
              <thead>
              <tr>
                <th>{% trans %}Name{% endtrans %}</th>
                <th>{% trans %}Description{% endtrans %}</th>
              </tr>
              </thead>
              <tbody>
                {% for col in data['collections']  %}
                {% if col.get('itemType') != 'record' %}
                <tr>
                  <td data-label="name">
                    <a  title="{{ col['title'] | striptags | truncate }}"
                      href="{{ data['collections_path'] }}/{{ col.id }}">
                      <span>{{ col['title'] | striptags | truncate }}</span></a>
                  </td>
                  <td data-label="description">
                    {{ col['description'] | striptags | truncate }}
                  </td>
                </tr>
                {% endif %}
                {% endfor %}
              </tbody>
            </table>
    </section>
{% endblock %}
