<!DOCTYPE html>
<html>
  <head>
    <title>ReDoc - {{ config['metadata']['identification']['title'] }}</title>
    <meta charset="utf-8"/>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link href="{{ config['server']['url'] }}/static/fonts_montserra_roboto.css" rel="stylesheet">
    <style>
      body {
        margin: 0;
        padding: 0;
      }
    </style>
  </head>
  <body>
    <redoc spec-url='{{ data['openapi-document-path'] }}?f=json'></redoc>
    <script src="{{ config['server']['url'] }}/static/redoc.standalone.js"> </script>
  </body>
</html>
